#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#include <errno.h>
#include <stdio.h>
#ifdef STDC_HEADERS
# include <stdlib.h>
# include <stddef.h>
#else
# ifdef HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif
#ifdef HAVE_STRING_H
# if !defined STDC_HEADERS && defined HAVE_MEMORY_H
#  include <memory.h>
# endif
# include <string.h>
#endif
#ifdef HAVE_SYS_RESOURCE_H
# include <sys/resource.h>
#endif
#ifdef TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# ifdef HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif
#include <sys/types.h>
#include <sys/wait.h>
#ifdef HAVE_SYS_PARAM_H
# include <sys/param.h>
#endif
#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif
#ifdef HAVE_GETOPT_LONG
# include <getopt.h>
#else
# include "getopt.h"
#endif
#if ! defined HAVE_GETHOSTNAME && defined HAVE_UNAME
# ifdef HAVE_SYS_UTSNAME_H
#  include <sys/utsname.h>
# endif
#endif

static const char version[] = VERSION;
static const char progname[] = PACKAGE;
static const char default_out_name[] = "getrusage.out";
enum {
  pid_prlen = sizeof(pid_t) * 3, /* we assume every byte uses 3
                                  * characters or less when printing */
};

static int
run_cmd(int argc, char *argv[]);
static int
parse_options(int argc, char *argv[], int *nonopt_start);
static int
print_cmd(int argc, char *argv[], FILE *fp);
static inline void
fprint_tag_timediff(FILE *fp, const char *tag, struct timeval t);

static struct rusage usage;
static int verbose = 0, outfn_insert_child_pid = 0, outfn_insert_node_name = 0,
  outfp_print_cmd = 0;
static char *summary_out_fn = NULL;
static pid_t child_pid;
static struct timeval start, end;

int
main(int argc, char *argv[])
{
  int status, rstatus, cmd_ofs = 1;
  if ((status = parse_options(argc, argv, &cmd_ofs)) != 0)
  {
    exit(status < 0?EXIT_FAILURE:status);
  }
  if (cmd_ofs >= argc)
  {
    fputs("no sub-command specified!\n", stderr);
    return EXIT_FAILURE;
  }
  if ((status = run_cmd(argc - cmd_ofs, argv + cmd_ofs)) < 0)
  {
    fputs("could not fork child\n", stderr);
    return EXIT_FAILURE;
  }
  if ((rstatus = getrusage(RUSAGE_CHILDREN, &usage) == -1))
  {
    perror("resource usage statistics unavailable");
  }
  else if (rstatus != 0)
  {
    fputs("an unknown error occurred\n", stderr);
    return EXIT_FAILURE;
  }
  else
  {
    FILE *summary_out = stderr;
    if (outfn_insert_child_pid || outfn_insert_node_name)
      do {
        char *last_dot_pos, *last_slash_pos;
        char *new_summary_out_fn = NULL;
        char node_name_res[MAXHOSTNAMELEN + 2] = "",
          child_pid_string[pid_prlen + 2] = "";
#if ! defined HAVE_GETHOSTNAME && defined HAVE_UNAME
        struct utsname system_information;
#endif
        if ((!summary_out_fn
             && !(summary_out_fn = strdup(default_out_name)))
            || !(new_summary_out_fn
                 = malloc(strlen(summary_out_fn) + 1 + pid_prlen
                          + 1 + MAXHOSTNAMELEN + 1 + 1)))
          {
            fprintf(stderr, "%s: opening file \"%s\" failed: %s\n", argv[0],
                    summary_out_fn, strerror(errno));
            break;
          }
        last_slash_pos = strrchr(summary_out_fn, '/');
        if (!last_slash_pos) last_slash_pos = summary_out_fn;
        last_dot_pos = strrchr(last_slash_pos, '.');
        if (!last_dot_pos)
          last_dot_pos = last_slash_pos + strlen(last_slash_pos);
        if (outfn_insert_child_pid)
          sprintf(child_pid_string, ".%u", (unsigned)child_pid);
        if (outfn_insert_node_name)
        {
          node_name_res[0] = '.';
#if defined HAVE_GETHOSTNAME
          if (gethostname(node_name_res + 1, MAXHOSTNAMELEN) != 0)
#elif defined HAVE_UNAME
          if (uname(&system_information) == 0)
            strncpy(node_name_res + 1, system_information.nodename,
                    MAXHOSTNAMELEN + 1);
#endif
            node_name_res[0] = '\0';
        }
        sprintf(new_summary_out_fn, "%*.*s%s%s%s",
                (int)(last_dot_pos - summary_out_fn),
                (int)(last_dot_pos - summary_out_fn),
                summary_out_fn, child_pid_string,
                node_name_res, last_dot_pos);
        free(summary_out_fn);
        summary_out_fn = new_summary_out_fn;
      } while (0);
    if (summary_out_fn)
    {
      if (!(summary_out = fopen(summary_out_fn, "w")))
      {
        fprintf(stderr, "%s: opening file \"%s\" failed: %s\n", argv[0],
                summary_out_fn, strerror(errno));
        summary_out = stderr;
      }
    }
    if (outfp_print_cmd)
      print_cmd(argc - cmd_ofs, argv + cmd_ofs, summary_out);
    end.tv_sec -= start.tv_sec + (end.tv_usec < start.tv_usec);
    end.tv_usec -= start.tv_usec - ((end.tv_usec < start.tv_usec)?1000000:0);
    fprint_tag_timediff(summary_out, "wall-clock time", end);
    fprint_tag_timediff(summary_out, "user time", usage.ru_utime);
    fprint_tag_timediff(summary_out, "sys  time", usage.ru_stime);
    fprintf(summary_out, "%s: %ldk\n", "max rss", usage.ru_maxrss);
    fprintf(summary_out, "%s: %ldkiBs\n", "integral shared memory size",
            usage.ru_ixrss);
    fprintf(summary_out, "%s: %ldkiBs\n", "integral unshared data size",
            usage.ru_idrss);
    fprintf(summary_out, "%s: %ldkiBs\n", "integral unshared stack size",
            usage.ru_isrss);
    fprintf(summary_out, "%s: %ld\n", "page reclaims", usage.ru_minflt);
    fprintf(summary_out, "%s: %ld\n", "page faults", usage.ru_majflt);
    fprintf(summary_out, "%s: %ld\n", "swaps", usage.ru_nswap);
    fprintf(summary_out, "%s: %ld\n", "block input operations",
            usage.ru_inblock);
    fprintf(summary_out, "%s: %ld\n", "block output operations",
            usage.ru_oublock);
    fprintf(summary_out, "%s: %ld\n", "messages sent", usage.ru_msgsnd);
    fprintf(summary_out, "%s: %ld\n", "messages received", usage.ru_msgrcv);
    fprintf(summary_out, "%s: %ld\n", "voluntary context switches",
            usage.ru_nvcsw);
    fprintf(summary_out, "%s: %ld\n", "involuntary context switches",
            usage.ru_nivcsw);
  }
  return status;
}

static inline void
fprint_tag_timediff(FILE *fp, const char *tag, struct timeval t)
{
    fprintf(fp, "%s: %ldm%3fs\n", tag,
            t.tv_sec/60, (t.tv_sec%60 * 1000000 + t.tv_usec)/1000000.0);
}

static void
getreltimeofday(struct timeval *tv)
{
#ifdef HAVE_GETTIMEOFDAY
  int status = gettimeofday(tv
# ifdef GETTIMEOFDAY_TWO_ARGS
                      , NULL
# endif
    );
  if (status == -1)
  {
    perror("time aquisition failed");
    exit(1);
  }
#endif
}

static void
print_usage(FILE *out, char * const argv0);
static void
print_short_usage(char * const argv0);

static int
parse_options(int argc, char *argv[], int *nonopt_start)
{
  /* parse_args_success is returned by the function
     0 indicates successful parsing of arguments without meeting an
     exit condition.
     >0 means parsing of args has failed
     <0 means parsing of args was successful but the supplied
     options need exit of program with status 0
  */
  int parse_args_status=0;

  static const struct option long_options[]=
    {
      /*  declare valid options */
      /*  the long options are the authorative way to customize
          program operation, not all might be implemented via single letter
          versions */
      /*  optname, has arg,  */
      { "output", required_argument, 0, 'o'},
      { "child-pid", no_argument, 0, 'p'},
      { "node-name", no_argument, 0, 'n'},
      { "print-cmd", no_argument, 0, 'c'},
      { "help", no_argument, 0, 'h'},
      { "verbose", no_argument, 0, 'v'},
      { "noverbose", no_argument, 0, 0},
      { "version", no_argument, 0, 'V'},
      { 0, 0, 0, 0}
    };

  int c;
  static const char short_options[]=":o:hV?vpcn";
  int option_index=1;

  optind = 1;
  while(!parse_args_status
        && optind < argc
        && argv[optind][0] == '-'
        && (c=getopt_long(argc, argv, short_options,
                          long_options, &option_index))!=-1)
    switch(c)
    {
    case 0:
      if (!strcmp(long_options[option_index].name, "noverbose"))
        verbose = 0;
      else
        parse_args_status=1;
      break;
    case 'h':
    case '?':
      print_usage(stdout, argv[0]);
      parse_args_status = -1;
      break;
    case 'v':
      verbose = 1;
      break;
    case 'V':
      printf("%s %s\n", progname, version);
      parse_args_status=-1;
      break;
    case 'o':
      if(!(summary_out_fn = realloc(summary_out_fn, strlen(optarg) + 1)))
      {
        perror("cannot use name for file");
        break;
      }
      strcpy(summary_out_fn, optarg);
      break;
    case 'p':
      outfn_insert_child_pid = 1;
      break;
    case 'n':
      outfn_insert_node_name = 1;
      break;
    case 'c':
      outfp_print_cmd = 1;
      break;
    default:
      parse_args_status=1; /* unknown option encountered */
      print_short_usage(argv[0]);
      break;
    }
  if(parse_args_status==0)
  {
    *nonopt_start = optind;
  }
  return parse_args_status;
}

static int
run_cmd(int argc, char *argv[])
{
  int status;
  if (verbose)
    print_cmd(argc, argv, stderr);
  if (argc < 1)
    return -1;
  getreltimeofday(&start);
  if ((child_pid = fork()) < 0)
    status = -1;
  else if (child_pid == 0)
  {
    /* child */
    execvp(argv[0], argv);
    _exit(127); /* execvp should not have returned */
  }
  else
  {
    while (waitpid(child_pid, &status, 0) < 0)
      if (errno != EINTR)
      {
        status = -1;
        break;
      }
  }
  getreltimeofday(&end);
  return status;
}

static void
print_usage(FILE *out, char * const argv0)
{
  fprintf(
    out, "%s %s, a program for resource summary printing\n"
    "Usage: %s [OPTION...] CMD [CMDOPTS...]\n"
    "Send bug-reports to: "PACKAGE_BUGREPORT"\n"
    "Mandatory arguments to long options are mandatory for short "
    "options too.\n\n"
    "  -V,  --version      display the version of %s\n"
    "  -v,  --verbose      print verbose progress information\n"
    "       --noverbose    don't print verbose progress information\n"
    "  -h,  --help         print this help.\n"
    "  -o,  --output=FILE  write to FILE instead of standard output.\n"
    "  -p,  --child-pid    add child process id to output FILE name.\n"
#if defined HAVE_GETHOSTNAME || defined HAVE_UNAME
    "  -n,  --node-name    add network identifier of host to output FILE name.\n"
#endif
    "  -c,  --print-cmd    print command executed to output\n",
    progname, version, argv0, progname
    );
}

static void
print_short_usage(char * const argv0)
{
  fprintf(
    stderr, "Usage: %s [OPTION]...\n\n"
    "Try \'%s --help\' for more options.\n",
    argv0, argv0
    );
}

static int
print_cmd(int argc, char *argv[], FILE *fp)
{
  int argl = 0, prsize, prsizesum = 0;
  static const char prprefix[] = "command and arguments:";
  if (fputs(prprefix, fp) == EOF)
    return -1;
  prsizesum = sizeof(prprefix);
  while(argv[argl])
    if ((prsize = fprintf(fp, " %s", argv[argl++])) == 0)
      return -1;
    else
      prsizesum += prsize;
  if (fputs("\n", fp) == EOF)
    return -1;
  return ++prsizesum;
}

